
/* asyncread.c - perform asynchronous reads              */
/* this program attempts to generate num_streams read    */
/* requests asynchronously then waits to collect them    */
/* through an aiowait() function call. This function     */
/* can be passed a timeval parameter if the user need    */
/* only wait to specified time. A parameter of NULL will */
/* get the process to wait indefinitely.                 */

#include <sys/types.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#ifdef __SOLARIS__
#include <sys/asynch.h>
#else
#include <aio.h>
#endif
#include <sys/time.h>
#include <time.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>

#include "global.h"
#include "timer.h"
#include "options.h"
#include "since.h"

// #define DEBUG

#define RESULT_OK       0
#define RESULT_IN_USE   1

#define TRUE    1
#define FALSE   0

int    how_big;
extern int block_size;
// char   buf[MAX_BUF_SIZE];

int aioReadMeg(char*, int);

struct streamdata {		/* data structure for each stream */
	int fileno;		/* The filenumber for this stream */
	int pos;		/* The position in the file       */
	int fd;			/* The file's file descriptor     */
        char buf[MAX_BUF_SIZE]; /* This stream's buffer.          */
	int size;		/* The file size.                 */
	int done;               /* True when this stream is done. */
#ifdef __SOLARIS__
        aio_result_t result;    /* The result of the aioread */
#else
        struct aiocb result;
#endif
} streamdata[MAX_STREAMS];

int main(int argc, char *argv[]) {
	int i = 0, j = 0;
	int rv;
	char *filename;
	int fd;
	int ctr = 0, errctr = 0;
	struct stat sbuf;
#ifdef __SOLARIS__
	aio_result_t* rp;
#else
	struct aiocb *rp;
#endif

	/* see options.c */
	extern int num_streams, bit_rate, num_blocks;

	/* see options.c */
        parseArgs(argc, argv);
        if(validateParams() != VALIDATION_SUCCESS) {
                printUsage();
                exit(1);
        }   

	// init_result();

	for (i = 0; i < num_streams; i++) {
		streamdata[i].fileno = i;
		streamdata[i].done = RESULT_OK;
		streamdata[i].pos = 0;
		filename = (char*)malloc(256);
	        if(filename == NULL) {
       			fprintf(stderr, "Can't malloc for filename.\n");
                	exit(1);
        	}
        
        	rv = snprintf(filename, 255, "bf/%2.2d",i);
        	if(rv != 5) {
                	fprintf(stderr, "Can't snprintf for filename.\n");
        	}

        
		fd = open(filename, O_RDONLY, NULL);
        	if(fd < 1) {
                	printf("open error: fd %d on %s\n",fd, filename);
			exit(1);
        	}
		free(filename);
		streamdata[i].fd = fd;

		rv = fstat(fd, &sbuf);
		if(rv != 0) {
			printf("fstat error: %d\n",rv);
			exit(1);
		}
		streamdata[i].size = sbuf.st_size;
#ifdef DEBUG		
		printf("Initialized stream: %d\n",i);
#endif
	}

	how_big = num_blocks;
        if(how_big > streamdata[0].size / block_size) {
                fprintf(stderr, "Can't read %d blocks from file!!\n", how_big);
                exit(1);
        }

#ifdef DEBUG
	printf("how_big: %d, block_size: %d, file_size: %d\n",
		how_big, block_size, how_big * block_size);
#endif
        
	/* See timer.c */
	init_timer(block_size, bit_rate);

        /* See since.c */
        init_since();

	for (i = 0; i < how_big; i++) {
                now_since();
		for(j = 0; j < num_streams; j++) {
#ifdef DEBUG
                	printf("Requesting chunk %d in stream %d.\n",i, j);
#endif
                	if(aioReadMeg(streamdata[j].buf, j) < 0) {
                       		fprintf(stderr, "Exiting because of stream %d.\n", j);
                        	exit(1);
                	}
#ifdef DEBUG
                	fprintf (stderr, "Stream %d block %d.\n", j, i);
#endif
		}

		for(j = 0; j < num_streams; j++) {
                        
#ifdef __SOLARIS__
                        rp = aiowait(NULL);
                        if(rp == (aio_result_t *)-1) {
                                fprintf(stderr, "Fatal: aiowait returned -1.");
                                exit(1);
                        }
#ifdef DEBUG
                        printf("return: %d, error: %d\n", rp->aio_return, rp->aio_errno);
#endif /* DEBUG */

#else /* __SOLARIS__ */
                        /* Need to: make sure we don't get two at once. */
                        rv = aio_waitcomplete(&rp, NULL);
                        if(rv == -1) {
                                perror("Fatal: aiowait returned -1");
                                exit(1);
                        }
#ifdef DEBUG
                        printf("return: %d\n", rv);
#endif /* DEBUG */
#endif /* __SOLARIS__ */
                        
                }

                for(j = 0; j < num_streams; j++) {
#ifdef __SOLARIS__
                        if(streamdata[j].result.aio_return != block_size) {
                                fprintf(stderr, "Fatal error: result (%d)not equal block_size in stream %d\n", streamdata[j].result.aio_return, j);
                                exit(1);
                        }
#endif /* __SOLARIS__ */
                        streamdata[j].pos += 1;
#ifdef DEBUG
                        printf("Chunk %d in stream %d returned from its read.\n",i, j);
#endif
                }
                since();
		++ctr;
		if(sleep_end_quantum_increment()) {
			++errctr;
		}
		printf("Period %d, %d errors\n",ctr,errctr);
        }
        end_since();
        printf("Total errors: %d\n", errctr);
	exit(0);
} /* end main */





int aioReadMeg(char *buf, int fn) {
        int rv;
	//streamdata* ns;
#ifndef __SOLARIS__
        struct aiocb iocb;
#endif /* not __SOLARIS__ */

#ifdef DEBUG
        fprintf(stderr, "Commencing read stream %d, buf loc %d, block size %d, offset %d\n",
                fn, (int)&buf, block_size, streamdata[fn].pos * block_size);
#endif
        //ns = (streamdata*)nextStream();
#ifdef __SOLARIS__
        rv = aioread(streamdata[fn].fd, buf, block_size,
                     (off_t)(streamdata[fn].pos * block_size), SEEK_SET,
                     &(streamdata[fn].result));
#else
        memset(&iocb, NULL, sizeof(struct aiocb));
        iocb.aio_nbytes = block_size;
        iocb.aio_fildes = streamdata[fn].fd;
        iocb.aio_offset = streamdata[fn].pos * block_size;
        iocb.aio_buf = buf;
        rv = aio_read(&iocb);
#endif /* __SOLARIS__ */
        if(rv != 0) {
                perror("Can't aioread");
                exit(1);
        }
#ifdef DEBUG
        fprintf(stderr, "Initiated aioread.\n");
#endif
        return rv;
}
