/* rrfiles.c
 *
 * $Id: rrfiles.c,v 1.9 2000/12/04 03:37:03 alderman Exp $
 *
 * round robin scheduling.
 */
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pthread.h>
#include <sys/time.h>
#include <unistd.h>

#include "global.h"
#include "options.h"
#include "timer.h"

struct streamdata {		/* data structure for each stream / thread */
	pthread_t t_tid;	/* the thread id */
	int fileno;		/* The filenumber for this stream */
	int pos;		/* The position in the file */
	int fd;			/* The file's file descriptor */
	int size;               /* The file size. */
	int done;               /* True when this thread is done. */
	char buf[NUMBYTES+1];   /* The thread's buffer. */
} threaddata[MAX_STREAMS];

/* Size of file in bytes; calculated from one of the files. */
int how_big;

pthread_cond_t sched_go;
pthread_mutex_t sched_mutex;
int currthread = 0;
int scheduling = 1;

pthread_t sched_tid;

void *readStream(void *fnp);
void *scheduler(void *idp);
int readMeg(char *buf, int fd);

struct timeval start_tv;

int main(int argc, char *argv[]) {
	int i = 0;
	int rv;
	struct stat sbuf;
	char *filename;

	pthread_attr_t attr;

	/* options.c */
	extern int block_size, num_streams;

	parseArgs(argc, argv);
	if(validateParams() != VALIDATION_SUCCESS) {
		printUsage();
		exit(1);
	}

        /* Calculate how_big */
	filename = malloc(256);
        if(filename == NULL) {
                fprintf(stderr, "Can't malloc for filename.\n");
                exit(1);
        }

        rv = snprintf(filename, 255, "bf/%2.2d",0);
        if(rv != 5) {
                fprintf(stderr, "Can't snprintf for filename.\n");
        }
	rv = stat(filename, &sbuf);
	how_big = sbuf.st_size / block_size;
	free(filename);

	rv = pthread_attr_init(&attr);
	if(rv) {
		perror("Can't init pthread attributes.");
		exit(1);
	}

//        rv = pthread_setdetachstate(&attr, PTHREAD_CREATE_UNDETACHED);
        if(rv) {
                perror("Can't set detached state");
                exit(1);
        }

        pthread_cond_init(&sched_go, NULL);
        pthread_mutex_init(&sched_mutex, NULL);

	for (i = 0; i < num_streams; i++) {
		threaddata[i].fileno = i;
		threaddata[i].done = 0;
		printf("Creating thread: %d\n",i);
		rv = pthread_create(&(threaddata[i].t_tid), 
					&attr, readStream, 
					(void *)&(threaddata[i].fileno));
		if(rv) {
			perror("Can't create thread");
			exit(1);
		}
	}

        rv = pthread_create(&sched_tid, &attr, scheduler, NULL);
	if(rv) {
                perror("Can't create scheduler thread");
                exit(1);
        }
        
	for(i = 0; i < MAX_STREAMS; i++) {
		printf("Waiting for thread: %d\n",i);
		pthread_join(threaddata[i].t_tid, NULL);
	}
        pthread_join(sched_tid, NULL);
	exit(0);
}

void *readStream(void *fnp) {
        int fd, i, rv;
	char *filename;
	char *buf;

	int fn = *(int *)fnp;

	buf = threaddata[fn].buf;

	printf("Reading stream: %d\n", fn);

        memset(buf, '\0', NUMBYTES);

	filename = malloc(256);
	if(filename == NULL) {
		fprintf(stderr, "Can't malloc for filename.\n");
		exit(1);
	}

	rv = snprintf(filename, 255, "bf/%2.2d",fn);
	if(rv != 5) {
		fprintf(stderr, "Can't snprintf for filename.\n");
	}

        fd = open(filename, O_RDONLY, NULL);
        if(fd < 1) {
                printf("error: fd %d on %s\n",fd, filename);
        }
	free(filename);

        for (i = 0; i < how_big; i++) {
                pthread_mutex_lock(&sched_mutex);
                while(scheduling || currthread != fn) {
                        pthread_cond_signal(&sched_go);
                        pthread_cond_wait(&sched_go,&sched_mutex);
                }
		printf("Reading chunk %d in stream %d.\n",i, fn);
                if(readMeg(buf, fd) < 0) {
                        fprintf(stderr, "Exiting with i = %d.\n", i);
                        exit(1);
                }
		fprintf (stderr, "Stream %d block %d.\n", fn, i);
                scheduling = 1;
                pthread_mutex_unlock(&sched_mutex);
        }
        pthread_mutex_lock(&sched_mutex);
        pthread_cond_signal(&sched_go);
        threaddata[fn].done = 1;
        pthread_mutex_unlock(&sched_mutex);
	return(0);
}

void *scheduler(void *idp) {
        // int *my_id = idp;
        int all_done = 0;
        int i, errctr = 0;

	extern int block_size, num_streams, bit_rate;
	
        init_timer(block_size, bit_rate);

	while(!all_done) {
                all_done = 1;
                pthread_mutex_lock(&sched_mutex);
                for(i = 0; i < num_streams; i++) {
                        printf("Scheduler unblocking thread %d\n",i);
                        /* Unblock threads in order */
                        if(!threaddata[i].done) {
                                all_done = 0;
                                currthread = i;
                                scheduling = 0;
                                pthread_cond_signal(&sched_go);

                                while(!scheduling) {
                                        pthread_cond_wait(&sched_go,&sched_mutex);
                                }
                        }
			if(sleep_end_quantum_increment()) {
				++errctr;
			}
			printf("Period %d, %d errors\n", i, errctr);
                }
                pthread_mutex_unlock(&sched_mutex);
        }
	return(0);
}

               

int readMeg(char *buf, int fd) {
        int curByte = 0;
        int rv;
        while(1) {
                rv = read(fd, buf, NUMBYTES - curByte);
                curByte += rv;
                if(rv > 0) {
                        // fprintf(stderr, "Read %d bytes.\n",rv);
                        if (curByte < NUMBYTES) {
                                continue;
                        } else {
				return curByte;
			}
                } else if (rv == 0) {
                        fprintf(stderr, "Read Zero bytes.\n");
                        return rv;
                } else {
                        fprintf(stderr, "Read returned negative: %d.\n", rv);
                        return rv;
                }
        }
        return rv;
}
